"""A Model"""

from __future__ import annotations

import logging
from collections import defaultdict
from collections.abc import Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, Self, Type

from dill import dump

from .._core._x import _X
from ..components.commodities.currency import Currency
from ..components.commodities.emission import Emission
from ..components.commodities.land import Land
from ..components.commodities.material import Material
from ..components.commodities.resource import Resource
from ..components.game.couple import Interact
from ..components.game.player import Player
# from ..components.graph.edge import Edge
# from ..components.graph.node import Node
from ..components.impact.categories import Economic, Environ, Social
from ..components.measure.unit import Unit
from ..components.operations.process import Process
from ..components.operations.storage import Storage
from ..components.operations.transport import Transport
from ..components.spatial.linkage import Linkage
from ..components.spatial.location import Location
from ..components.temporal.modes import Modes
from ..components.temporal.periods import Periods
from ..components.temporal.scales import TemporalScales
from ..dimensions.impact import Impact
from ..dimensions.problem import Problem
from ..dimensions.space import Space
from ..dimensions.system import System
from ..dimensions.time import Time
from ..library.aliases import aspect_aliases
from ..library.instructions import costing_commodity, costing_operation
from ..library.recipes import (capacity_sizing, economic, environmental,
                               free_movement, inventory_sizing, operating,
                               social, trade, usage)
from ..modeling.parameters.instruction import Instruction
from ..modeling.variables.control import Control
from ..modeling.variables.recipe import Recipe
from ..modeling.variables.states import Consequence, State, Stream
from .ations.graph import Graph
from .ations.program import Program
from .ations.scenario import Scenario

logger = logging.getLogger("energia")
logger.setLevel(logging.INFO)

ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
formatter = logging.Formatter("%(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)


if TYPE_CHECKING:
    from enum import Enum
    from typing import DefaultDict

    from .._core._component import _Component
    from ..components.commodities.commodity import Commodity
    from ..modeling.indices.domain import Domain
    from ..modeling.indices.sample import Sample
    from ..modeling.variables.aspect import Aspect

    BalanceType = DefaultDict[
        Commodity,
        DefaultDict[Location | Linkage, DefaultDict[Periods, list[Aspect]]],
    ]


@dataclass
class Model:
    """
    An abstract representation of an energy system.

    :param name: Name of the Model. Defaults to m.
    :type name: str
    :param init: List of functions to initialize the Model. Defaults to None.
    :type init: list[Callable], optional
    :param default: True if some default objects should be declared.
    :type default: bool
    :param capacitate: True if process capacities should be determined to bound operations.
    :type capacitate: bool

    :ivar added: List of added objects to the Model.
    :vartype added: list[str]
    :ivar update_map: maps component type to representation and collection.
    :vartype update_map: dict
    :ivar time: time representation of the Model.
    :vartype time: Time
    :ivar space: spatial representation of the Model.
    :vartype space: Space
    :ivar impact: impact representation of the Model.
    :vartype impact: Impact
    :ivar tree: feasible region (Decision-Making) of the Model.
    :vartype tree: Tree
    :ivar graph: Graph (Network) of the Model.
    :vartype graph: Graph
    :ivar system: System (Resource Task Network) of the Model.
    :vartype system: System
    :ivar program: Mathematical (mixed integer) program of the Model.
    :vartype program: Program
    :ivar conversions: List of Balances in the Model.
    :vartype conversions: list[Conversion]
    :ivar convmatrix: Conversion matrix of the Model.
    :vartype convmatrix: dict[Process, dict[Resource, int | float | list]]
    :ivar modes_dict: Dictionary mapping Bind objects to Modes.
    :vartype modes_dict: dict[Bind, Modes]
    :siunits_set: True if SI units have been set.
    :vartype siunits_set: bool
    :ivar cookbook: Recipes to create Aspects.
    :vartype cookbook: dict[str, Recipe]
    :ivar directory: Map of attribute names to recipes for creating them.
    :vartype directory: dict[str, dict[str, Recipe]]
    :ivar classifiers: List of classifiers for the Model.
    :vartype classifiers: list[Enum]
    :ivar grb: Dictionary which tells you what aspects of resource have GRB {loc: time: []} and {time: loc: []}.
    :vartype grb: DefaultDict[Commodity,DefaultDict[Location | Linkage, DefaultDict[Periods, list[Aspect]]]]
    :ivar dispositions: Dictionary which tells you what aspects of what component have been bound at what location and time.
    :vartype dispositions: dict[Aspect, dict[Commodity | Process | Storage | Transport, dict[Location | Linkage, dict[Periods, list[Aspect]]]]]
    :ivar maps: Maps of aspects to domains.
    :vartype maps: dict[Aspect, dict[Domain, dict[str, list[Domain]]]]
    :ivar maps_report: Maps of aspects to domains for reporting variables.
    :vartype maps_report: dict[Aspect, dict[Domain, dict[str, list[Domain]]]]


    :raises ValueError: If an attribute name already exists in the Model.
    """

    name: str = "m"
    init: list[Callable[[Self]]] | None = None
    default: bool = True
    capacitate: bool = False

    def __post_init__(self):

        self.reserved_names = []

        # what components have been added to the model
        self.added: list[str] = []
        # map of what representation and collection within that representation
        # an object of a particular type belongs to

        # --------------------------------------------------------------------
        # * Component Mapping to Dimension and Collection
        # --------------------------------------------------------------------
        # Dimensions in brackets
        self.familytree = {
            # * I Temporal (Time):
            # 1.  Periods (Periods) generates a bespoke discretization.
            Periods: ("time", "periods"),
            # 2. Modes discrete options in the same time
            Modes: ("time", "modes"),
            # * II Spatial (Space):
            # Spatial representation (Space).
            # 1. a bespoke discretization.
            Location: ("space", "locations"),
            # 2. link between them
            Linkage: ("space", "linkages"),
            # * III Streams (System):
            # All are Commodity derived:
            # 1. Money (Currency)
            Currency: ("system", "currencies"),
            # 2. Land (Land) resource
            Land: ("system", "lands"),
            # 3. Emission (Emission) resource
            Emission: ("system", "emissions"),
            # Resource is a general Commodity
            # These are Resource subsets
            # 1. Material (Material) used to setup processes
            Material: ("system", "materials"),
            # 2. etc. societal (Jobs), etc (Etc).
            Resource: ("system", "resources"),
            # * IV Operations (System):
            # 1. A production operation (Process) involves conversion of resources
            Process: ("system", "processes"),
            # 2. A transport operation (Transport) which describes a task in the system
            # that involves transporting resources from
            #    one location to another.
            Transport: ("system", "transports"),
            # 3. A storage operation (Storage) stores (charges)
            # and retrieves (discharge) resources
            Storage: ("system", "storages"),
            # *V Indicators (Consequence):
            # scales a stream and projects onto a common metric
            # categories include
            Environ: (
                "impact",
                "environment",
            ),
            Social: (
                "impact",
                "society",
            ),
            Economic: ("impact", "economy"),
            # * VI Game Components
            # To model Competition
            Player: ("game", "players"),
            Interact: ("game", "interacts"),
            # * VII Problem Aspects
            # The problem at hand
            # 1. to control the volume of streams
            Control: ("problem", "controls"),
            # 2. movement
            Stream: ("problem", "streams"),
            # 3. size, quantity
            State: ("problem", "states"),
            # 4. consequence
            Consequence: ("problem", "consequences"),
        }

        self.reserved_names += zip(*self.familytree.values())

        # --------------------------------------------------------------------
        # * Dimensions or Representation
        # --------------------------------------------------------------------

        # * I Dimensions
        # * 1. Time with Periods and Modes
        self.time = Time(self)
        # * 2. Space with Locations and Linkages
        self.space = Space(self)
        # * 3. Impact with Indicator categories
        self.impact = Impact(self)
        # * 4. System (Resource Task Network)
        self.system = System(self)

        # * II Representations
        # * 1. Graph with Edges and Nodes
        self.graphs = [Graph(self)]
        # * 2. Problem at hand
        self.problems = [Problem(self)]
        # * 3 Mathematical Program of mpMINLP subclass
        self.programs = [Program(model=self)]
        # * 4 Scenario, the parameter set or uncertainty realization
        self.scenarios = [Scenario(model=self)]

        # shorthand
        self._ = self.program

        # --------------------------------------------------------------------
        # * Attributes Inherited from Dimensions or Representations
        # --------------------------------------------------------------------
        # Start with patterened
        self.program_attrs = [
            "constraint",
            "function",
            "variable",
            "parameter",
            "theta",
        ]
        # word -> words and word_sets
        self.program_attrs += [
            w + s for w in self.program_attrs for s in ['s', '_sets']
        ]
        self.program_attrs += ["solution", "formulation", "evaluation"]
        # word -> n_word
        self.program_attrs += ['n_' + w for w in self.program_attrs]
        self.program_attrs += [
            "index_sets",
            "indices",
            "objectives",
            "parameter_sets",
            "X",
        ]

        self.reserved_names += self.program_attrs

        # properties that can be called by model
        # these never get set
        _program_matrices = [
            "A",
            "B",
            "C",
            "F",
            "G",
            "H",
            "CrA",
            "CrB",
            "NN",
            "A_with_NN",
            "B_with_NN",
            "Z",
            "P",
        ]

        self.reserved_names += _program_matrices

        self.properties = {
            "horizon": self.time,
            "network": self.space,
            "indicators": self.impact,
            "operations": self.system,
            "aspects": self.problem,
            "domains": self.problem,
            **{i: self.program for i in _program_matrices},
        }

        # --------------------------------------------------------------------
        # * Default Components
        # --------------------------------------------------------------------
        # if any of these attributes are called,
        # or an exiting one is returned
        self.default_components = {
            "l": self._l0,
            "l0": self._l0,
            "t0": self._t0,
            "t": self._t0,
            "money": self._cash,
        }

        self.graph_components = ["edges", "nodes"]

        # --------------------------------------------------------------------
        # * Books of Maps Between:
        # --------------------------------------------------------------------
        # * matching_aspect -> Recipe
        self.cookbook: dict[str, Recipe] = {}
        # * parameter_name -> parameter_handling_instruction
        self.manual: dict[str, Instruction] = {}
        # * already_defined_user_input_attr -> matching_aspect
        self.registry: dict[str, Aspect] = {}
        # * user_input_attr -> matching_aspect -> Recipe
        self.directory: dict[str, dict[str, Recipe]] = {}
        # * collection -> dimension
        # derived from familytree
        self.ancestry = {
            collection: dimension for dimension, collection in self.familytree.values()
        }
        # * collection to component
        # derived from familytree
        self.ilk = {
            collection: component
            for component, (_, collection) in self.familytree.items()
        }

        # --------------------------------------------------------------------
        # * Constraint Ledger
        # --------------------------------------------------------------------

        # Dictionary which tells you what aspects of resource
        # have been set in what location and time

        # * General Resource Balances
        self.balances: BalanceType = defaultdict(
            lambda: defaultdict(lambda: defaultdict(list))
        )
        # Dictionary which tells you what aspects of what component
        # have been bound at what location and time

        # * Sample Dispositions
        self.dispositions: dict[
            Aspect,
            dict[
                Commodity | Process | Storage | Transport,
                dict[Location | Linkage, dict[Periods, list[Aspect]]],
            ],
        ] = {}

        # * Drawn Maps
        self.maps: dict[Aspect, dict[Domain, dict[str, list[Domain]]]] = {}
        self.maps_report: dict[Aspect, dict[Domain, dict[str, list[Domain]]]] = {}

        # * Generated Modes
        self.modes_dict: dict[Sample, Modes] = {}

        # * Conversion Matrix
        self.convmatrix: dict[Process, dict[Resource, int | float | list]] = {}

        # --------------------------------------------------------------------
        # * Measurement Related
        # --------------------------------------------------------------------
        self.units: list[Unit] = []
        # if SI units have been set
        self.siunits_set: bool = False

        # --------------------------------------------------------------------
        # * Model Classification
        # --------------------------------------------------------------------
        self.classifiers: dict[str, list[Enum]] = {
            "uncertainty": [],
            "structure": [],
            "scale": [],
            "paradigm": [],
        }

        # --------------------------------------------------------------------
        # * Model Initialization
        # --------------------------------------------------------------------
        # functions are passed and initialized on self
        if not self.init:
            self.init = []

        if self.default:
            self.init += [
                # Recipes
                capacity_sizing,
                operating,
                inventory_sizing,
                free_movement,
                trade,
                economic,
                environmental,
                social,
                usage,
                aspect_aliases,
                # Instructions
                costing_operation,
                costing_commodity,
            ]

        for func in self.init:
            func(self)

    # -------------------------------------------------------------------
    # * Active Representation
    # -------------------------------------------------------------------
    @property
    def problem(self) -> Problem:
        """The active problem"""
        return self.problems[-1]

    @property
    def graph(self) -> Graph:
        """The active graph"""
        return self.graphs[-1]

    @property
    def program(self) -> Program:
        """The active program"""
        return self.programs[-1]

    @property
    def scenario(self) -> Scenario:
        """The active scenario"""
        return self.scenarios[-1]

    # -------------------------------------------------------------------
    # * Onboard Component and Send to Family
    # -------------------------------------------------------------------

    def update(
        self,
        name: str,
        value: _X,
        represent: str,
        collection: str,
        aspects: list[str] | None = None,
    ):
        """Update the Model with a new value

        :param name: Name of the value to be added
        :type name: str
        :param value: Value to be added
        :type value: X
        :param represent: Representation to which the value belongs
        :type represent: str
        :param collection: Collection within the representation to which the value belongs
        :type collection: str
        :param aspects: Aspects to be added to the value, defaults to None
        :type aspects: list[str], optional
        """

        value.name = name
        # every component is handed the model
        value.model = self

        if name in self.added:
            # do not allow overriding of components
            # throw error if name already exists
            raise ValueError(f"{name} already defined")
            # added is the list of all components that have been added to the model
        self.added.append(name)

        # if not subset:
        #     # ignore subsets

        model_set: list = getattr(getattr(self, represent), collection)
        # the set that needs to be updated
        model_set.append(value)

        # update the index set for index elements
        if collection in [
            "resources",
            "currencies",
            "lands",
            "emissions",
            "materials",
            "processes",
            "storages",
            "transits",
            "locations",
            "linkages",
        ]:
            setattr(
                self.program, collection, getattr(self.program, collection) | value.I
            )

        # set aspect samples on the components
        if aspects:
            for asp in aspects:
                aspect = getattr(self, asp)

                setattr(value, asp, aspect(value))

                if aspect.neg is not None:
                    setattr(value, aspect.neg.name, aspect.neg(value))

    # -------------------------------------------------------------------
    # * Birthing Procedures and Setting Aliases
    # -------------------------------------------------------------------
    # These take an action on attribute inputs

    def Recipe(
        self,
        name: str,
        kind: Type[Aspect],
        primary_type: tuple[Type[_Component]] | Type[_Component],
        label: str = "",
        latex: str = "",
        add: str = "",
        add_latex: str = "",
        add_kind: Type[Aspect] | None = None,
        sub: str = "",
        sub_latex: str = "",
        sub_kind: Type[Aspect] | None = None,
        neg: str = "",
        neg_latex: str = "",
        neg_label: str = "",
        bound: str = "",
        ispos: bool = True,
        nn: bool = True,
    ):
        """Creates a Recipe and updates recipes

        :param name: Name of the aspect
        :type name: str
        :param kind: type of the aspect
        :type kind: Type[Aspect]
        :param primary_type: type of primary component
        :type primary_type: tuple[Type[_Component]] | Type[_Component]
        :param label: label for the aspect. Defaults to ''.
        :type label: str, optional
        :param latex: LaTeX representation for the aspect. Defaults to None.
        :type latex: str, optional
        :param add: add control variable. Defaults to ''.
        :type add: str, optional
        :param add_latex: LaTeX representation for the add aspect. Defaults to ''.
        :type add_latex: str, optional
        :param add_kind: type of the add aspect. Defaults to None.
        :type add_kind: Type[Aspect], optional
        :param sub: sub control variable. Defaults to ''.
        :type sub: str, optional
        :param sub_latex: LaTeX representation for the sub aspect. Defaults to ''.
        :type sub_latex: str, optional
        :param sub_kind: type of the sub aspect. Defaults to None.
        :type sub_kind: Type[Aspect], optional
        :param neg: name of the negative aspect. Defaults to ''.
        :type neg: str, optional
        :param neg_latex: LaTeX representation for the negative aspect. Defaults to ''.
        :type neg_latex: str, optional
        :param neg_label: label for the negative aspect. Defaults to ''.
        :type neg_label: str, optional
        :param bound: name of the bound aspect. Defaults to ''.
        :type bound: str, optional
        :param ispos: whether the aspect is positive. Defaults to True.
        :type ispos: bool, optional
        :param nn: whether the aspect is non-negative. Defaults to True.
        :type nn: bool, optional
        """
        if name in self.cookbook:
            logger.warning("⛔ Overriding existing recipe: %s ⛔", name)

        self.cookbook[name] = Recipe(
            name=name,
            kind=kind,
            label=label or name,
            add=add,
            sub=sub,
            bound=bound,
            ispos=ispos,
            nn=nn,
            primary_type=primary_type,
            latex=latex,
        )

        if add:
            self.Recipe(
                name=add,
                kind=add_kind or sub_kind or Control,
                primary_type=primary_type,
                label=add_latex or add,
                ispos=True,
                nn=True,
                latex=latex or add,
            )

        if sub:
            self.Recipe(
                name=sub,
                kind=sub_kind or add_kind or Control,
                primary_type=primary_type,
                label=sub_latex or sub,
                ispos=False,
                nn=True,
                latex=latex or sub,
            )

        if neg:
            neg_recipe = Recipe(
                name=neg,
                kind=kind,
                label=neg_label,
                ispos=not ispos,
                nn=nn,
                primary_type=primary_type,
                latex=neg_latex or neg,
            )
            self.cookbook[neg] = neg_recipe

    def alias(self, *names: str, of: str):
        """Set aspect aliases

        :param names: Names of the aliases
        :type names: str
        :param to: Name of the aspect to which the aliases point
        :type to: str
        """
        _add = dict.fromkeys(list(names), {of: self.cookbook[of]})
        self.directory = {**self.directory, **_add}

    def Instruction(
        self,
        name: str,
        kind: Type[_Component],
        deciding: str,
        depending: str,
        default: str,
        label: str = "",
        latex: str = "",
    ):
        """Creates an Instruction and updates the manual

        :param deciding: Name of the deciding aspect
        :type deciding: str
        :param depending: Name of the depending aspect
        :type depending: str
        :param default: Name of the default component
        :type default: str
        :param label: Label for the parameter. Defaults to ''.
        :type label: str, optional
        :param latex: LaTeX representation for the parameter. Defaults to ''.
        :type latex: str, optional
        """

        self.manual[name] = Instruction(
            name=name,
            kind=kind,
            deciding=deciding,
            depending=depending,
            default=default,
            label=label,
            latex=latex,
        )

    # ------------------------------------------------------------------------
    # * Easy Birthing of Components
    # ------------------------------------------------------------------------

    def declare(self, what: Type[_X], names: list[str]):
        """Declares objects conveniently

        :param what: Type of object to be created
        :type what: Type[X]
        :param names: Names of the objects to be created
        :type names: list[str]
        """
        for i in names:
            setattr(self, i, what())

    def Link(
        self,
        source: Location,
        sink: Location,
        dist: float = 0,
        bi: bool = False,
    ):
        """
        Link two Locations

        :param source: Source Location
        :type source: Location
        :param sink: Sink Location
        :type sink: Location
        :param dist: Distance between the Locations. Defaults to None.
        :type dist: float | Unit, optional
        :param bi: Whether the linkage is bidirectional. Defaults to False.
        :type bi: bool, optional
        """
        if source - sink:
            # if source and sink are already linked
            raise ValueError(
                f"A link already defined between {source} and {sink}.\n"
                "For multiple linkages with different attributes, use model.named_link = Link(...)",
            )

        link = Linkage(source=source, sink=sink, dist=dist, bi=bi, auto=True)
        setattr(self, f"{source.name}-{sink.name}", link)

    def TemporalScales(self, discretizations: list[int], names: list[str]):
        """
        This is an easy way to define multiple time periods (scales)

        :param discretizations: List of discretizations for the temporal scale.
        :type discretizations: list[int]
        :param names: Names of the discretizations. Defaults to [t<i>] for each discretization.
        :type names: list[str], optional
        """
        # set the root period:
        setattr(self, names[-1], Periods())
        # pick up the period that was just created
        # use it as the root
        root = self.periods[-1]
        discretizations = list(reversed(discretizations))

        names = list(reversed(names[:-1]))

        if discretizations[-1] != 1:
            discretizations.append(1)
            names.append("t0")

        for disc, name in zip(discretizations, names):
            setattr(self, name, disc * root)
            root = self.periods[-1]

    def Modes(self, size: int, sample: Sample):
        """
        This is an easy way to define modes within a period

        :param size: Number of modes to create
        :type size: int
        :param name: Name of the modes. Defaults to "modes".
        :type name: str, optional
        """
        modes = Modes(size=size, sample=sample, n=len(self.modes))
        periods = sample.domain.periods or self.time.horizon
        setattr(self, f'_{periods}{len(periods.modes)}', modes)
        periods.modes.append(modes)
        return modes

    # ------------------------------------------------------------------------
    # * Illustrations from Different Perspectives
    # ------------------------------------------------------------------------

    # * I Mathematical
    def show(
        self,
        descriptive: bool = False,
        categorical: bool = True,
        category: str = "",
    ):
        """
        Pretty print the Model

        :param descriptive: Whether to show descriptive information. Defaults to False.
        :type descriptive: bool, optional
        :param categorical: Whether to group by category. Defaults to True.
        :type categorical: bool, optional
        :param category: If provided, shows only this category. Defaults to None.
        :type category: str, optional
        """
        self.program.show(descriptive, categorical=categorical, category=category)

    # * II Graphical
    def draw(self, variable: Aspect | Sample | None = None, n_sol: int = 0):
        """
        Draw the solution for a variable

        :param variable: Variable to draw. Defaults to None.
        :type variable: Aspect | Sample | None, optional
        :param n_sol: Solution number to draw. Defaults to 0.
        :type n_sol: int, optional

        """
        if variable is not None:
            self.program.draw(variable=variable.V(), n_sol=n_sol)

        else:
            self.program.draw(n_sol=n_sol)

    # * III Solution
    def output(self, n_sol: int = 0, slack: bool = True, compare: bool = False):
        """Solution"""
        return self.program.output(n_sol=n_sol, slack=slack, compare=compare)

    # ------------------------------------------------------------------------
    # * Solution Prep, Generation, and  Handling
    # ------------------------------------------------------------------------
    def locate(self, *operations: Process | Storage):
        """Locate operations in the network

        :param operations: Operations to locate
        :type operations: Process | Storage
        """
        self.network.locate(*operations)

    # * Optimization
    def solve(
        self,
        using: Literal[
            "combinatorial",
            "combinatorial_parallel",
            "combinatorial_parallel_exp",
            "graph",
            "graph_exp",
            "graph_parallel",
            "graph_parallel_exp",
            "combinatorial_graph",
            "geometric",
            "geometric_parallel",
            "geometric_parallel_exp",
        ] = "combinatorial",
    ):
        """
        Solve the multiparametric program

        :param using: The solving method to use. Defaults to "combinatorial".
        :type using: Literal[
            "combinatorial",
            "combinatorial_parallel",
            "combinatorial_parallel_exp",
            "graph",
            "graph_exp",
            "graph_parallel",
            "graph_parallel_exp",
            "combinatorial_graph",
            "geometric",
            "geometric_parallel",
            "geometric_parallel_exp",
        ], optional
        """

        self.program.solve(using=using)

    # * Solution evaluation
    def eval(
        self, *theta_vals: float, n_sol: int = 0, roundoff: int = 4
    ) -> list[float]:
        """
        Evaluate the objective function at given theta values

        :param theta_vals: values for the parametric variables
        :type theta_vals: float
        :param n_sol: solution number to evaluate, defaults to 0
        :type n_sol: int, optional
        :param roundoff: number of decimal places to round off to, defaults to 4
        :type roundoff: int, optional

        :return: list of objective function values
        :rtype: list[float]
        """
        return self.program.eval(*theta_vals, n_sol=n_sol, roundoff=roundoff)

    # * Saving
    def save(self, as_type: str = "dill"):
        """Save the Model to a file"""
        if as_type == "dill":
            with open(self.name + ".energia", "wb") as f:
                dump(self.solution, f)
        else:
            raise ValueError(f"Unknown type {as_type} for saving the model")

    # ------------------------------------------------------------------------
    # * Default Components
    # ------------------------------------------------------------------------

    def _t0(self, size: int = 1) -> Periods:
        """Return a default period

        :param size: Size of the period. Defaults to 0.
        :type size: int, optional

        :return: Periods object
        :rtype: Periods
        """

        if not self.periods:
            # if no periods exits yet, make the horizon
            self.t0 = Periods()

        if size > 1:
            setattr(self, f"t{len(self.periods)}", self.horizon / size)

        return self.periods[-1]

        # or create a default period

    def _l0(self) -> Location:
        """Return a default location"""
        self.l0 = Location(label="l")
        return self.l0

    def _cash(self) -> Currency:
        """Return a default currency"""
        if self.currencies:
            return self.currencies[0]
        self.cash = Currency(label="$")
        return self.cash

    # -------------------------------------------------------------------
    # * Attribute Setting and Getting
    # -------------------------------------------------------------------

    def __setattr__(self, name, value):

        if isinstance(value, (str, dict, list, bool)) or value is None:
            # if value is a string, dict, list or bool
            # set the attribute to the value
            super().__setattr__(name, value)
            return

        if isinstance(value, TemporalScales):
            self.TemporalScales(value.discretizations, value.names)

            return

        if isinstance(value, Unit):
            value.name = name
            self.units.append(value)

        # map to representation and collection
        for cls, updates in self.familytree.items():
            if isinstance(value, cls):
                # for args in updates:
                self.update(name, value, *updates)
                break

        # Special linkage instructions
        if isinstance(value, Linkage):

            self.space.sources.append(value.source)
            self.space.sinks.append(value.sink)

            if value.bi:
                # if bidirectional, set the reverse linkage
                # also ensures that all linakges go in one direction only
                rev = value.rev()
                setattr(self, rev.name, rev)

        super().__setattr__(name, value)

    def __getattr__(self, name):
        # Only called when attribute does not exist

        # if something like t, t0 is called
        # just return a default component
        # t/t0, l/l0, cash, money
        # this will not intefere with the setting of
        # attributes what this name
        if name in self.default_components:
            component = self.default_components[name]()
            return component

        # Inherits collections based on ancestry
        if name in self.ancestry:
            dimension = getattr(self, self.ancestry[name])
            collection = getattr(dimension, name)
            setattr(self, name, collection)
            return collection

        # Program attributes
        if name in self.program_attrs:
            collection = getattr(self.program, name)
            setattr(self, name, collection)
            return collection

        # properties from dimensions and representations
        if name in self.properties:
            return getattr(self.properties[name], name)

        # already declare and mapped to aspect
        if name in self.registry:
            return self.registry[name]

        if name in self.manual:
            return self.manual[name]

        # Recipe for defining aspects
        if name in self.cookbook:
            recipe = self.cookbook[name]

            aspect = recipe.kind(**recipe.args)
            setattr(self, name, aspect)

            return aspect
        # maps many attribute names to aspects
        if name in self.directory:
            # if this is an attribute being called for the first time
            recipe = self.directory[name]
            aspect_name = list(recipe.keys())[0]

            if aspect_name in self.added:
                # if same aspect is called by a different name
                return getattr(self, aspect_name)

            # these are the arguments for the aspect
            recipe = recipe[aspect_name]
            aspect = recipe.kind(**recipe.args)

            setattr(self, aspect_name, aspect)

            self.registry[name] = aspect

            return aspect

        raise AttributeError(
            f"{self} has no '{name}'",
        )

    # ---------------------------------------------------------------
    # * Call to Initialize using functions and Hashing
    # ---------------------------------------------------------------

    def __call__(self, *funcs: Callable[[Self]]):
        """Set functions on the model

        These can include default units

        """

        for f in funcs:
            f(self)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)



"""Process"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ...utils.decorators import timer
from .operation import Operation

logger = logging.getLogger("energia")

if TYPE_CHECKING:
    from ..spatial.location import Location
    from ..temporal.periods import Periods


class Process(Operation):
    """
    Process converts one Resource to another Resource at some Location

    :param label: An optional label for the component. Defaults to None.
    :type label: str, optional
    :param citations: An optional citation or description for the component. Defaults to None.
    :type citations: str | list[str] | dict[str, str | list[str]], optional

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar conversion: Operational conversion associated with the operation. Defaults to None.
    :vartype conversion: Conversion, optional
    :ivar _conv: True if the operational conversion has been set. Defaults to False.
    :vartype _conv: bool
    :ivar fab: Material conversion associated with the operation. Defaults to None.
    :vartype fab: Conversion, optional
    :ivar _fab_balanced: True if the material conversion has been balanced. Defaults to False.
    :vartype _fab_balanced: bool
    :ivar locations: Locations at which the process is balanced. Defaults to [].
    :vartype locations: list[Location]
    :ivar charges: If the Process is Storage charging. Defaults to None.
    :vartype charges: Storage, optional
    :ivar discharges: If the Process is Storage discharging. Defaults to None.
    :vartype discharges: Storage, optional
    """

    def __init__(self, *args, label: str = "", citations: str = "", **kwargs):

        Operation.__init__(self, *args, label=label, citations=citations, **kwargs)

        # at which locations the process is balanced
        # Note that we do not need a conversion at every temporal scale.
        # once balanced at a location for a particular time,
        # if time != horizon, the individual streams are summed up anyway
        self.locations: list[Location] = []

    @property
    def spaces(self) -> list[Location]:
        """Locations at which the process is balanced"""
        return self.locations

    @timer(logger, kind="production")
    def write_production(self, space_times: list[tuple[Location, Periods]]):
        """Write the production constraints for the process"""

        if not self.production:
            logger.warning(
                "%s: Production not defined, no Constraints generated",
                self.name,
            )
            return

        # This makes the production consistent
        # check conv_test.py in tests for examples
        self.production.balancer()

        #! PWL
        # if self.conversion.pwl:
        # if there are piece-wise linear conversions
        # here we assume that the same resources appear in all piece-wise segments
        # this is a reasonable assumption for conversion in processes
        # but not if process modes involve different resources

        # TODO:
        # make the statement eff = [conv[res] for conv in self.conversion.values()]
        # into try
        # if that fails, create a consistent dict, see:
        # {0: {r1: 10, r2: -5}, 1: {r1: 8, r2: -4, r3: -2}}
        # transforms to {0: {r1: 10, r2: -5, r3: 0}, 1: {r1: 8, r2: -4, r3: -2}}
        # the r3: 0 will ensure that r3 is considered in all modes
        # the zero checks will prevent unnecessary constraints
        # there is a problem though, because I am only checking for the elements in the first dict
        # in the multi conversion dict

        #     conversion = self.balance[list(self.balance)[0]]

        # else:

        for space, time in space_times:

            if space in self.locations:
                # if the process is already balanced for the space , Skip
                continue

            self.production.write(space, time)

            # update the locations at which the process exists
            self.locations.append(space)
            self.space_times.append((space, time))

        return self, self.locations


# for res, par in self.production.items():
#     # set, the conversion on the resource

#     #! Repurpose
#     # setattr(res, self.name, self)

#     # now there are two cases possible
#     # the parameter (par) is positive or negative
#     # if positive, the resource is expended
#     # if negative, the resource is produced
#     # also, the par can be an number or a list of numbers

#     # insitu resource (produced and expended within the system)
#     # do not initiate a grb so we need to run a check for that first

#     if res in self.model.balances:
#         time = time_checker(res, location, time)
#         _ = self.model.balances[res].get(location, {})

#     eff = par if isinstance(par, list) else [par]

#     if eff[0] < 0:
#         # Resources are consumed (expendend by Process) immediately
#         rhs = res.expend(self.operate, location, time)
#         eff = [-e for e in eff]
#     else:
#         # Production — may occur after lag
#         lag_time = self.lag.of if self.lag else time
#         rhs = res.produce(self.operate, location, lag_time)

#     opr = self.operate(location, time)

#! PWL
# because of using .balancer(), expend/produce are on same temporal scale

# if self.conversion.pwl:

#     eff = [conv[res] for conv in self.balance.values()]

#     if eff[0] < 0:
#         eff = [-e for e in eff]

#     if not self.conversion.modes_set:
#         # this is setting the bin limits for piece wise linear conversion
#         # these are written bound to capacity generally
#         # but here we pause that binding and bind operate to explicit limits
#         self.model.operate.bound = None

#         _ = opr == dict(enumerate(self.balance.keys()))

#         # reset capacity binding
#         self.model.operate.bound = self.model.capacity

#         modes = self.model.modes[-1]
#         self.conversion.modes_set = True

#     else:
#         modes = self.conversion.modes
#         modes.bind = self.operate
#         self.conversion.modes_set = True

#     opr = opr(modes)

#     rhs = rhs(modes)

# _ = opr(modes)[rhs(modes)] == eff

# else:
# _ = opr[rhs] == eff



"""Commodity Base Class"""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Self

from ..._core._component import _Component
from ...modeling.parameters.conversion import Conversion

if TYPE_CHECKING:
    from ..measure.unit import Unit


class Commodity(_Component, Mapping):
    """
    A commodity, can be a material, chemical, energy, etc.

    :param label: Label of the commodity, used for plotting. Defaults to None.
    :type label: str, optional
    :param name: Name of the commodity, used for indexing. Defaults to None.
    :type name: str, optional
    :param basis: Unit basis of the commodity. Defaults to None.
    :type basis: Unit, optional


    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str
    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar conversions: List of conversions associated with the commodity. Defaults to [].
    :vartype conversions: list[Conversion]
    :ivar insitu: If the commodity only exists insitu, i.e., does not scale any domains
    :vartype insitu: bool, optional
    """

    def __init__(
        self, basis: Unit | None = None, label: str = "", citations: str = "", **kwargs
    ):
        self.basis = basis
        _Component.__init__(self, label=label, citations=citations, **kwargs)

        # list of conversions associated with the commodity
        self.conversions: list[Conversion] = []

        # flag indicates whether the commodity is produced and expended insitu only
        self.insitu = False

    @property
    def balance(self) -> dict[Conversion | Self, int | float]:
        """Conversion"""
        # if no conversion is set, return 1.0
        if not self.conversions:
            return {self: 1.0}
        # if there is a conversion, return its parameter in all tasks its
        # associated with
        return {task: task.balance[self] for task in self.conversions}

    def __setattr__(self, name, value):

        if isinstance(value, Conversion):
            # update conversions
            self.conversions.append(value)

        super().__setattr__(name, value)

    def __mul__(self, other: int | float) -> Conversion:
        # multiplying a number with a resources gives conversion
        # math operations with conversions form the balance in tasks
        return Conversion.from_balance({self: other})

    def __rmul__(self, other: int | float) -> Conversion:
        # reverse multiplication
        return self * other

    def __add__(self, other: Conversion | Commodity) -> Conversion:
        return Conversion.from_balance({self: 1, **other})

    def __neg__(self) -> Conversion:
        # just multiply by -1
        return self * -1

    def __sub__(self, other: Conversion | Self):
        if isinstance(other, Commodity):
            # if another resource is subtracted
            # give it the parameter -1
            return self + -1 * other
        if isinstance(other, Conversion):
            # if another conversion is subtracted, update the balance
            return Conversion.from_balance(
                {
                    self: 1,
                    **{
                        res: (
                            -1 * par
                            if isinstance(par, (int, float))
                            else [-i for i in par]
                        )
                        for res, par in other.items()
                    },
                }
            )

    def __truediv__(self, other: int | float):
        # treat division as multiplication by the inverse
        return self * (1 / other)

    def __eq__(self, other: Conversion | Self):
        if isinstance(other, Conversion):
            conv = self + other
            # set itself as base
            conv.resource = self
            return conv

        if isinstance(other, int | float):
            conv = Conversion(resource=self, hold=other)

            return conv

        return super().__eq__(other)

    def __getitem__(self, _):
        return 1

    def __iter__(self):
        return iter([self])

    def __len__(self):
        return 1



"""Bound Parameter Input"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..._core._name import _Name

if TYPE_CHECKING:
    from ...components.temporal.periods import Periods


class Value(_Name):
    """Input Value"""

    def __init__(
        self,
        value: (
            float
            | int
            | list[float]
            | list[int]
            | tuple[float]
            | tuple[int]
            | tuple[list[int | float]]
            | None
        ) = None,
        periods: Periods | None = None,
    ):
        self.value = value
        self.periods = periods
        _Name.__init__(self, label="")
        if isinstance(self.value, (float, int)):
            self.name = f"{self.value}/{self.periods}"
        else:
            self.name = f"Φ/{self.periods}"



"""Impact Indicator"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..._core._component import _Component

if TYPE_CHECKING:
    from ...dimensions.impact import Impact


class Indicator(_Component):
    """
    Impact Indicator

    :param label: An optional label for the component. Defaults to None.
    :type label: str, optional
    :param citations: An optional citation or description for the component. Defaults to None.
    :type citations: str | list[str] | dict[str, str | list[str]], optional

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    """

    def __init__(self, label: str = "", citations: str = "", **kwargs):

        _Component.__init__(self, label=label, citations=citations, **kwargs)

    @property
    def consequence(self) -> Impact:
        """Impact object"""
        return self.model.impact



"""Inherited _Index class"""

from __future__ import annotations

from abc import ABC, abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING

from ._hash import _Hash

if TYPE_CHECKING:
    from gana import I as Idx
    from gana import Prg
    from gana.sets.constraint import C

    from ..modeling.indices.domain import Domain
    from ..modeling.variables.aspect import Aspect
    from ..represent.model import Model


class _X(ABC, _Hash):
    """
    A component (`x`) that functions as an index in the mathematical program.

    :param label: An optional label for the component. Defaults to None.
    :type label: str, optional
    :param citations: Citation for the component. Defaults to None.
    :type citations: str, optional

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str
    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]

    .. note:
        - `name` and `model` are set when the component
           is assigned as a Model attribute.
        - `constraints` and `domains` are populated as the program is built.
    """

    def __init__(
        self,
        label: str = "",
        citations: str = "",
    ):
        self.label = label
        self.citations = citations
        # the model
        self.model: Model | None = None
        # name is given by the model
        self.name: str = ""

        # constraint pnames associated with the component
        self.constraints: list[str] = []
        # domains associated with the component
        self.domains: list[Domain] = []
        # aspects associated with the component with domains
        self.aspects: dict[Aspect, list[Domain]] = {}

    @cached_property
    def program(self) -> Prg:
        """Mathematical program"""
        if self.model is None:
            raise ValueError(f"{type(self)} needs to be assign as Model attribute")
        return self.model.program

    @property
    def cons(self) -> list[C]:
        """Constraints"""
        # this gets the actual constraint objects from the program
        # based on the pname (attribute name) in the program
        return [getattr(self.program, c) for c in self.constraints]

    @property
    @abstractmethod
    def I(self) -> Idx:
        """gana index set"""

    def show(self, descriptive=False, category: str = ""):
        """Pretty print the component"""
        if category:
            for c in self.cons:
                if c.category and c.category == category:
                    c.show(descriptive)

        else:
            for c in self.cons:
                c.show(descriptive)



"""Pre-set Parameters"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..components.commodities.resource import Commodity
from ..components.operations.operation import Operation
from ..components.operations.storage import Stored

if TYPE_CHECKING:
    from ..represent.model import Model


def costing_operation(model: Model):
    """Sets costing parameters for operations"""

    model.Instruction(
        name="capex",
        kind=Operation,
        deciding="capacity",
        depending="spend",
        default="money",
        latex="Capex",
        label="Capital Expenditure",
    )
    model.Instruction(
        name="opex",
        kind=Operation,
        deciding="operate",
        depending="spend",
        default="money",
        latex="Opex",
        label="Operational Expenditure",
    )

    model.Instruction(
        name="invcapex",
        kind=Stored,
        deciding="invcapacity",
        depending="spend",
        default="money",
        latex="Capex",
        label="Capital Expenditure (Storage)",
    )

    model.Instruction(
        name="inventory_cost",
        kind=Stored,
        deciding="inventory",
        depending="spend",
        default="money",
        latex=r"Cost^{inv}",
        label="Inventory Cost (Storage)",
    )


def costing_commodity(model: Model):
    """Sets costing parameters for resources"""

    model.Instruction(
        name="price",
        kind=Commodity,
        deciding="consume",
        depending="spend",
        default="money",
        latex=r"Price",
        label="Resource Price",
    )



"""System"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from .._core._dimension import _Dimension

if TYPE_CHECKING:
    from ..components.game.couple import Interact
    from ..components.game.player import Player

@dataclass
class Game(_Dimension):
    """
    System representation as a Resource Task Network (RTN)
    All resources and tasks are attached to this object

    :param model: Model to which the representation belongs.
    :type model: Model

    :ivar name: Name of the dimension, generated based on the class and model name.
    :vartype name: str
    :ivar players: List of players. Defaults to [].
    :vartype players: list[Player]
    :ivar couples: List of couples. Defaults to [].
    :vartype couples: list[Couple]
    """

    def __post_init__(self):
        # ------------Decision-Makers----------------------
        self.players: list[Player] = []
        self.interacts: list[Interact] = []

        _Dimension.__post_init__(self)



"""Balance"""

from __future__ import annotations

import logging
from functools import cached_property
from operator import is_
from typing import TYPE_CHECKING, Self

from ..._core._hash import _Hash
from ...components.operations.storage import Stored
from ...utils.decorators import timer

logger = logging.getLogger("energia")

if TYPE_CHECKING:
    from gana.sets.constraint import C

    from ..._core._x import _X
    from ...components.spatial.linkage import Linkage
    from ...components.spatial.location import Location
    from ..indices.domain import Domain
    from ..variables.aspect import Aspect


class Balance(_Hash):
    """Performs a general commodity balance

    :param aspect: Aspect to which the constraint is applied
    :type aspect: Aspect
    :param domain: Domain over which the aspect is defined
    :type domain: Domain
    :param label: Label for the constraint. Defaults to "".
    :type label: str

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar program: The program to which the constraint belongs.
    :vartype program: Program
    :ivar grb: The general resource balance dictionary.t
    :vartype grb: dict[Commodity, dict[Location, dict[Periods, list[Aspect]]]]
    """

    def __init__(self, aspect: Aspect, domain: Domain):
        self.aspect = aspect
        self.domain = domain

        self._handshake()

        self.write()

    def write(self) -> bool | None:
        """Writes the stream balance constraint"""

        if self._check_existing():
            return False

        self._name = f"{self.commodity}_{self.space}_{self.time}_grb"

        if isinstance(self.commodity, Stored):
            stored = True

        else:
            stored = False
            if self.balances[self.commodity][self.space]:
                # If a GRB exists at a lower temporal order, append to that
                lower_times = [
                    t
                    for t in self.balances[self.commodity][self.space]
                    if t > self.time
                ]

                if lower_times:
                    _ = self.aspect(self.commodity, self.space, lower_times[0]) == True

        if not self.existing_aspects:
            # this checks whether a general self.commodity balance has been defined
            # for the self.commodity in that space and self.time

            return self._create_constraint(stored)

        # -add aspect to GRB if not added already ----

        return self._update_constraint(stored, getattr(self.program, self._name))

    @cached_property
    def space(self) -> Location | Linkage | None:
        """Location or Linkage of the constraint"""
        _space = (
            self.domain.linkage.source
            if self.aspect.sign == -1 and self.domain.linkage
            else (
                self.domain.linkage.sink
                if self.domain.linkage
                else self.domain.location
            )
        )

        if (
            _space.isin is not None
            and self.balances[self.commodity][_space.isin][self.time]
        ):

            return _space.isin
        return _space

    @cached_property
    def name(self) -> str:
        """Name of the constraint"""
        return f"{self.aspect.name}{self.domain}"

    @cached_property
    def sign(self) -> float:
        """Returns the aspect"""
        return self.aspect.sign

    @property
    def existing_aspects(self):
        """Exisiting Aspects"""
        return self.balances[self.commodity][self.space][self.time]

    @timer(logger, "balance-init")
    def _create_constraint(self, stored: bool) -> Domain:
        """
        Creates a new GRB constraint

        :param stored: If the commodity is stored
        :type stored: bool

        :returns: Domain in which the constraint was created
        :rtype: Domain
        """

        # The reason this is avoided is:
        # inv(t) - inv(t-1)  = #charge_eff*p_charge - #charge_eff*p_dcharge
        # for one time period, this becomes
        # inv(t) = #charge_eff*p_charge - #charge_eff*p_dcharge
        # if inv cost is not provided, this can become unbounded
        # or bounded to the max capacity
        if stored and self.aspect.name == "inventory":

            if len(self.time) == 1:
                return False

            # TODO: potential bug here
            # this avoids writing inv_ntw_t - inv_ntw_t-1 = 0
            # the need for this check is absurd
            # it ensures that inventory balances are not created for the network directly
            # when location-wise inventory balances can be written
            # this is a temporary fix and needs to be generalized
            if self.model.network.has and self.space.isnetwork:
                # if the location is a network, we cannot have storage
                return False

            # if inventory is being add to GRB
            lagged_domain = self.domain.change({"lag": -1 * self.time, "periods": None})
            cons_grb = -self(*self.domain).V() + self(*lagged_domain).V() == 0

        else:
            cons_grb = (
                self(*self.domain).V() == 0
                if self.aspect.ispos
                else -self(*self.domain).V() == 0
            )

        cons_grb.categorize("Balance")

        setattr(
            self.program,
            self._name,
            cons_grb,
        )
        self._inform()

        return self.domain

    @timer(logger, "balance-update")
    def _update_constraint(
        self,
        stored: bool,
        cons_grb: C,
    ) -> bool:
        """
        Updates an existing GRB constraint with the new aspect

        :param stored: If the commodity is stored
        :type stored: bool
        :param cons_grb: The existing GRB constraint
        :type cons_grb: C

        :returns: If the constraint was updated
        :rtype: bool
        """
        if stored and self.aspect.name == "inventory":

            if len(self.time) == 1:
                return False

            # if inventory is being add to GRB
            lagged_domain = self.domain.change({"lag": -1 * self.time, "periods": None})

            setattr(
                self.program,
                self._name,
                cons_grb + self(*lagged_domain).V() - self(*self.domain).V(),
            )
        else:
            setattr(
                self.program,
                self._name,
                (
                    cons_grb + self(*self.domain).V()
                    if self.aspect.ispos
                    else cons_grb - self(*self.domain).V()
                ),
            )

        self._inform()

        return self.domain, self.aspect

    def _check_existing(self) -> bool:
        """Checks if the balance constraint already exists"""
        if (
            (not self.samples and self.commodity)
            or (self.aspect(self.commodity, self.time) not in self.existing_aspects)
            or (self.commodity.insitu)
        ):
            return False
        return True

    def _inform(self):
        """
        Updates the constraints in all the indices of self.domain
        Add constraint name to aspect
        """
        self.domain.inform_indices(self._name)

        if self._name not in self.aspect.constraints:
            self.aspect.constraints.append(self._name)

        # update the GRB aspects
        self.existing_aspects.append(self)

    def _handshake(self):
        """Borrow attributes from aspect and domain"""
        self.model = self.aspect.model
        self.program = self.model.program
        self.balances = self.model.balances

        self.commodity = self.domain.commodity
        self.samples = self.domain.samples
        self.time = self.domain.time

    def __eq__(self, other: Self):
        return is_(self.aspect, other.aspect) and self.domain == other.domain

    def __call__(self, *index: _X):
        """Returns the variable for the aspect at the given index"""
        return self.aspect(*index)



"""A General Operation"""

from __future__ import annotations

import logging
from abc import abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING

from ..._core._component import _Component
from ...modeling.parameters.conversion import Conversion
from ...utils.decorators import timer

logger = logging.getLogger("energia")

if TYPE_CHECKING:
    from ...modeling.indices.sample import Sample
    from ...modeling.variables.aspect import Aspect
    from ..commodities.resource import Resource
    from ..spatial.linkage import Linkage
    from ..spatial.location import Location
    from ..temporal.lag import Lag
    from ..temporal.periods import Periods


class Operation(_Component):
    """A General Operation

    :param label: An optional label for the component. Defaults to None.
    :type label: str, optional
    :param citations: An optional citation or description for the component. Defaults to None.
    :type citations: str | list[str] | dict[str, str | list[str]], optional

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar conversion: Operational conversion associated with the operation. Defaults to None.
    :vartype conversion : Conversion, optional
    :ivar fab: Material conversion associated with the operation. Defaults to None.
    :vartype fab: Conversion, optional
    :ivar _fab_balanced: True if the material conversion has been balanced. Defaults to False.
    :vartype _fab_balanced: bool

    """

    def __init__(
        self,
        *args,
        # operaing_aspect: str,
        label: str = "",
        citations: str = "",
        **kwargs,
    ):
        _Component.__init__(self, label=label, citations=citations, **kwargs)

        self.production = Conversion(
            operation=self,
            aspect='operate',
            add="produce",
            sub="expend",
            attr_name="production",
        )

        self.construction = Conversion(
            operation=self,
            aspect='capacity',
            add="dispose",
            sub="use",
            attr_name="construction",
            use_max_time=True,
        )

        self.conversions = args
        self.space_times: list[tuple[Location | Linkage, Periods]] = []

    @property
    @abstractmethod
    def spaces(self) -> list[Location | Linkage]:
        """Locations at which the process is balanced"""

    @cached_property
    def capacity_aspect(self) -> Aspect:
        """Capacity Aspect"""
        return getattr(self.model, 'capacity')

    @cached_property
    def operate_aspect(self) -> Aspect:
        """Operate Aspect"""
        return getattr(self.model, 'operate')

    @property
    def operate_sample(self) -> Sample:
        """Operate Sample"""
        return getattr(self, 'operate')

    @property
    def capacity_sample(self) -> Sample:
        """Capacity Sample"""
        return getattr(self, 'capacity')

    @property
    def basis(self) -> Resource:
        """Base resource"""
        return self.production.resource

    @property
    def balance(self) -> dict[Resource, int | float]:
        """Conversion of commodities"""
        return self.production.balance

    @property
    def lag(self) -> Lag:
        """Lag of the process"""
        return self.production.lag

    def write_production(
        self,
        space_times: list[tuple[Location | Linkage, Periods]],
    ):
        """write conversion constraints for the operation"""

    @timer(logger, kind="construction")
    def write_construction(
        self,
        space_times: list[tuple[Location | Linkage, Periods]],
        # fabrication: dict[Resource, int | float | list[int | float]],
    ):
        """write fabrication constraints for the operation"""

        self.construction.balancer()

        for location, time in space_times:
            self.construction.write(location, time)

        return self, (l for l, _ in space_times)

    @timer(logger, kind='assume-capacity')
    def _check_capacity_bound(self, space: Location | Linkage):
        """Check if capacity is bounded in space"""

        if self not in self.capacity_aspect.bound_spaces:
            # ensure that the bound_spaces dict is initialized
            self.capacity_aspect.bound_spaces[self] = {"ub": [], "lb": []}

        if space not in self.capacity_aspect.bound_spaces[self]["ub"]:
            # check if operational capacity has been bound
            # this is not a check, this generates a constraint
            _ = self.capacity_sample(space, self.horizon) == True

            return self, space, self.horizon

        return False

    @timer(logger, kind='assume-operate')
    def _check_operate_bound(self, space: Location | Linkage):
        """Check if operate is bounded in space"""
        if self not in self.operate_aspect.bound_spaces:
            # ensure that the bound_spaces dict is initialized
            self.operate_aspect.bound_spaces[self] = {"ub": [], "lb": []}

        if space not in self.operate_aspect.bound_spaces[self]["ub"]:
            # check if operate has been bound
            # if not just write opr_{pro, space, horizon} <= capacity_{pro, space, horizon}

            if (
                self in self.operate_aspect.dispositions
                and space in self.operate_aspect.dispositions[self]
            ):
                time = min(self.operate_aspect.dispositions[self][space])
            else:
                time = self.horizon
            _ = self.operate_sample(space, time) <= 1

            return self, space, time

        return False

    @timer(logger, kind='locate')
    def locate(self, *spaces: Location | Linkage):
        """Locate the process"""

        if not spaces:
            spaces = (self.model.network,)

        # get location, time tuples where operation is defined
        space_times: list[tuple[Location | Linkage, Periods]] = []
        for space in spaces:

            self._check_capacity_bound(space)

            self._check_operate_bound(space)

            # check if the process is being operated at the location
            for d in self.operate_aspect.domains:
                if d.space == space:
                    space_time = (space, d.time)
                    if space_time not in space_times:
                        space_times.append(space_time)

        self.write_production(space_times)

        if self.construction is not None:
            self.write_construction(self.space_times)

        return self, spaces

    def __call__(
        self, resource: Resource | Conversion, lag: Lag | None = None
    ) -> Conversion:
        """Conversion is called with a Resource to be converted"""
        self.production.resource = resource
        if lag:
            return self.production(resource, lag)
        return self.production(resource)

    def __setattr__(self, name, value):

        if name == "model" and value is not None:
            for conv in self.conversions:
                conv.operation = self

            if len(self.conversions) == 1:
                self.production += self.conversions[0]
                self.production.resource = self.conversions[0].resource

        super().__setattr__(name, value)



"""Hashing Base Class"""


class _Hash:

    def __str__(self):
        return getattr(self, 'name')

    def __repr__(self):
        return getattr(self, 'name')

    def __hash__(self):
        return hash(getattr(self, 'name'))

    def __init_subclass__(cls):
        cls.__repr__ = _Hash.__repr__
        cls.__hash__ = _Hash.__hash__



"""Storage - Stashes Resource to Withdraw Later"""

from __future__ import annotations

import logging
from functools import cached_property
from typing import TYPE_CHECKING

from ..._core._component import _Component
from ...modeling.parameters.conversion import Conversion
from ...utils.decorators import timer
from ..commodities.resource import Resource
from .process import Process

logger = logging.getLogger("energia")

if TYPE_CHECKING:
    from gana.sets.constraint import C

    from ...components.temporal.periods import Periods
    from ...modeling.indices.sample import Sample
    from ...modeling.variables.aspect import Aspect
    from ..spatial.location import Location


# * Storage is made up of three components:


# Resource in Storage
class Stored(Resource):
    """Resource in Storage"""

    def __init__(self, *args, **kwargs):
        Resource.__init__(self, *args, **kwargs)

        # self.inv_of: Resource | None = None


# A charging process to convert Resource into Stored
class Charge(Process):
    """Process that Charges Storage"""

    def __init__(self, storage: Storage, *args, **kwargs):

        self.storage = storage
        super().__init__(*args, **kwargs)


# A discharging process to convert Stored back into Resource
class Discharge(Process):
    """Process that Discharges Storage"""

    def __init__(self, storage: Storage, *args, **kwargs):

        self.storage = storage
        super().__init__(*args, **kwargs)


class Storage(_Component):
    """
    Storage is a container for three main elements:
    1. A Stored resource
    2. A charge Process that puts the resource into storage
    3. A discharge Process that withdraws the resource from storage

    :param basis: Unit basis of the component. Defaults to None.
    :type basis: Unit, optional
    :param label: An optional label for the component. Defaults to None.
    :type label: str, optional
    :param citations: An optional citation or description for the component. Defaults to None.
    :type citations: str | list[str] | dict[str, str | list[str]], optional
    :param store: The resource to be stored. Defaults to None.
    :type store: Resource, optional

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str
    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar stored: The resource being stored. Defaults to None.
    :vartype stored: Stored, optional
    :ivar conversion: Operational conversion associated with the storage. Defaults to None.
    :vartype conversion: Conversion, optional
    :ivar _conv: True if the operational conversion has been set. Defaults to False.
    :vartype _conv: bool
    :ivar charge: The charge process associated with the storage.
    :vartype charge: Process
    :ivar discharge: The discharge process associated with the storage.
    :vartype discharge: Process
    :ivar locations: List of locations where the storage is located. Defaults to [].
    :vartype locations: list[Location]
    """

    def __init__(
        self,
        *args,
        label: str = "",
        citations: str = "",
        **kwargs,
    ):

        _Component.__init__(self, label=label, citations=citations, **kwargs)

        # Charging, Discharging, and Stored Resource (Inventory)
        self.charge: Charge | None = None
        self.discharge: Discharge | None = None
        self.stored: Stored | None = None

        # prevents repeated
        self._birthed = False

        self.locations: list[Location] = []

        self.conversions = args

        self.construction = Conversion(
            operation=self,
            aspect='invcapacity',
            add="dispose",
            sub="use",
            attr_name="construction",
            use_max_time=True,
        )

    @cached_property
    def space_times(self) -> list[tuple[Location, Periods]]:
        """List of location, time tuples where storage is located"""
        return self.charge.space_times

    @cached_property
    def capacity_aspect(self) -> Aspect:
        """Reports invcapacity as aspect"""
        return getattr(self.model, 'invcapacity')

    @cached_property
    def inventory_aspect(self) -> Aspect:
        """Reports inventory as aspect"""
        return getattr(self.model, 'inventory')

    @property
    def capacity(self) -> Sample:
        """Reports invcapacity as capacity"""
        return self.stored.invcapacity

    @property
    def setup(self) -> Sample:
        """Reports invsetup as setup"""
        return self.stored.invsetup

    @property
    def dismantle(self) -> Sample:
        """Reports invdismantle as dismantle"""
        return self.stored.invdismantle

    @property
    def inventory(self) -> Sample:
        """Inventory of the stored resource"""
        return self.stored.inventory

    @property
    def basis(self) -> Resource:
        """Base resource"""
        return self.discharge.production.basis

    @property
    def storage_cost(self) -> Sample:
        """Cost of storing the resource"""
        return self.inventory[self.model._cash().spend]

    @property
    def cons(self) -> list[C]:
        """Constraints"""
        # This overwrites the Component cons property
        # this gets the actual constraint objects from the program
        # based on the pname (attribute name) in the program
        return (
            [getattr(self.program, c) for c in self.constraints]
            + self.charge.cons
            + self.discharge.cons
            + self.stored.cons
        )

    @timer(logger, kind='assume-capacity', level=logging.INFO)
    def _check_capacity_bound(self, space: Location) -> bool:
        """Check if the storage capacity has been bound at that location"""

        if self not in self.capacity_aspect.bound_spaces:
            # ensure that the bound spaces dict exists
            self.capacity_aspect.bound_spaces[self.stored] = {"ub": [], "lb": []}

        if space not in self.capacity_aspect.bound_spaces[self.stored]["ub"]:
            # check if the storage capacity has been bound at that location
            # Note: this is not a check, this generates a constraint
            _ = self.capacity(space, self.horizon) == True

            return self, space, self.horizon

        return False

    @timer(logger, kind='assume-inventory', level=logging.INFO)
    def _check_inventory_bound(self, space: Location) -> bool:
        """Check if the storage inventory is capacity bound at that location"""
        if self.stored not in self.inventory_aspect.bound_spaces:
            _ = self.inventory_aspect(self.stored) == True

        if space not in self.inventory_aspect.bound_spaces[self.stored]["ub"]:
            # check if the storage inventory has been bound at that location
            try:
                times = list(
                    [
                        t
                        for t in self.model.balances[self.stored.inv_of][space]
                        if self.model.balances[self.stored.inv_of][space][t]
                    ],
                )
            except KeyError:
                times = []
            # write the conversion balance at
            # densest temporal scale in that space
            if times:
                time = min(times)
            else:
                time = self.horizon

            # if not just write opr_{pro, loc, horizon} <= capacity_{pro, loc, horizon}
            _ = self.inventory(space, time) <= 1
            return self, space, time

        return False

    @timer(logger, kind="construction")
    def write_construction(
        self,
        space_times: list[tuple[Location, Periods]],
        # fabrication: dict[Resource, int | float | list[int | float]],
    ):
        """write fabrication constraints for the operation"""

        self.construction.balancer()

        for location, time in space_times:
            self.construction.write(location, time)

        return self, (l for l, _ in space_times)

    @timer(logger, kind='locate')
    def locate(self, *spaces: Location):
        """Locate the storage"""
        # update the locations at which the storage exists

        # get location, time tuples where operation is defined
        for space in spaces:

            self._check_capacity_bound(space)

            self._check_inventory_bound(space)

        # locate the charge and discharge processes
        self.charge.locate(*spaces)
        self.discharge.locate(*spaces)

        if self.construction is not None:
            self.write_construction(self.space_times)

        return self, spaces

    def _birth_constituents(
        self,
        charging_args: dict | None = None,
        discharging_args: dict | None = None,
        storage_args: dict | None = None,
    ):
        """Birth the constituents of the storage component"""
        if not self._birthed:
            self.stored = Stored(**storage_args if storage_args else {})

            self.charge = Charge(storage=self, **charging_args if charging_args else {})

            self.discharge = Discharge(
                storage=self, **discharging_args if discharging_args else {}
            )

            # Set them on the model
            setattr(self.model, f"{self.name}.charge", self.charge)
            setattr(self.model, f"{self.name}.discharge", self.discharge)
            setattr(self.model, f"{self.name}.stored", self.stored)

        self._birthed = True

    def _split_attr(self):
        """Splits the parameters dictionary into charging, discharging and storage parameters"""

        _charging_args = {}
        _discharging_args = {}
        _storage_args = {}

        for attr, param in self.parameters.items():
            split_attr = attr.split("_")

            _attr = split_attr[0]

            if _attr == "charge":
                _charging_args["_".join(split_attr[1:])] = param
            elif _attr == "discharge":
                _discharging_args["_".join(split_attr[1:])] = param
            else:
                if attr[:3] == "inv":
                    _storage_args[attr] = param
                else:
                    # if there is no inv prefix.
                    _storage_args["inv" + attr] = param
        # reset parameters to empty
        # none of these go on Storage itself
        self.parameters = {}

        return _charging_args, _discharging_args, _storage_args

    def __setattr__(self, name, value):

        object.__setattr__(self, name, value)

        if name == "model" and value is not None:

            self._birth_constituents(*self._split_attr())

            if self.conversions:

                # if len(self.conversions) > 1:
                #     self.modes =

                for conversion in self.conversions:
                    if not isinstance(conversion, int | float):
                        conversion.operation = self

                if len(self.conversions) == 1:
                    conversion = self.conversions[0]

                    if conversion.hold is not None:
                        _ = self(conversion.resource) == conversion.hold

        super().__setattr__(name, value)

    def __call__(self, resource: Stored | Conversion):
        """Conversion is called with a Resource to be converted"""

        self._birth_constituents()

        _ = self.charge(self.stored) == -resource

        self.discharge.production.expect = self.stored

        self.stored.inv_of = resource

        return self.discharge(resource)



"""Decorators for functions"""

import logging
import time
from functools import wraps


def once(func):
    """Ensures the function is executed only once"""

    def wrapper(*args, **kwargs):
        if not hasattr(wrapper, "has_run"):
            wrapper.has_run = True
            return func(*args, **kwargs)

    return wrapper


def timer(
    logger: logging.Logger,
    kind=None,
    level=logging.INFO,
):
    """
    Logs execution time and optionally shows a full computation using function arguments and result.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start = time.time()
            # returns the result if successful, else False
            result = func(*args, **kwargs)
            elapsed = time.time() - start

            if result is not False:

                if kind == 'balance-update':

                    msg = f"⚖   Updated {result[0].commodity} balance with {result[1]}{result[0]}"

                elif kind == 'balance-init':

                    msg = f"⚖   Initiated {result.commodity} balance in ({result.space}, {result.time})"

                elif kind == 'map':
                    msg = f"🧭  Mapped {(result[1] - result[2])[0]} for {result[0]} {result[1]} ⟺ {result[2]}"

                elif kind == 'bind':
                    sample = result[0]
                    aspect = sample.aspect
                    domain = sample.domain
                    rel = result[1]

                    if rel == "ub":
                        rel = "≤"
                    elif rel == "lb":
                        rel = "≥"
                    else:
                        rel = "="

                    msg = f"🔗  Bound [{rel}] {domain.primary} {aspect} in ({domain.space}, {domain.time})"

                elif kind == 'assume-capacity':
                    msg = f"💡  Assumed {result[0]} capacity unbounded in ({result[1]}, {result[2]})"

                elif kind == 'assume-operate':
                    msg = f"💡  Assumed {result[0]} operate bounded by capacity in ({result[1]}, {result[2]})"

                elif kind == 'assume-inventory':
                    msg = f"💡  Assumed {result[0]} inventory bounded by capacity in ({result[1]}, {result[2]})"

                elif kind == 'locate':
                    msg = f"🌍  Located {result[0]} in {', '.join([str(s) for s in result[1]])}"

                elif kind == 'production':
                    msg = f"🏭  Operating streams introduced for {result[0]} in {', '.join([str(s) for s in result[1]])}"

                elif kind == 'construction':
                    msg = f"🏗   Construction streams introduced for {result[0]} in {', '.join([str(s) for s in result[1]])}"

                else:
                    msg = f"  Executed {func.__name__}"

                logger.log(
                    level,
                    f"{msg:<75} ⏱ {elapsed:.4f} s",
                )

        return wrapper

    return decorator



"""Aliases for aspects"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..represent.model import Model


def _create_default_aliases():
    _default_aliases = {
        'capacity': [
            'size',
            'volume',
            'throughput',
            'potential',
            'output',
            'cap',
        ],
        'setup': [
            'install',
            'build',
            'construct',
            'commission',
        ],
        'dismantle': [
            'remove',
            'take_apart',
            'decommission',
            'disassemble',
            'break_down',
            'minimize',
        ],
        'operate': ['run', 'control'],
        'rampup': ['ramp_up', 'boost', 'accelerate', 'expand', 'scale_up'],
        'rampdown': [
            'reduce',
            'decrease',
            'slow_down',
            'scale_down',
        ],
        'produce': ['generate', 'manufacture', 'create', 'fabricate', 'yield'],
        'expend': ['useup', 'use_up', 'utilize'],
        'ship_in': ['import', 'deliver_to', 'bring_in', 'transport_in'],
        'ship_out': ['send', 'export', 'dispatch', 'deliver', 'transport_out'],
        'inventory': ['inv', 'stock', 'storage', 'reserve', 'buffer'],
        'consume': ['take', 'absorb', 'deplete', 'procure', 'obtain'],
        'release': [
            'free',
            'deploy',
            'liberate',
            'give_off',
            'exhale',
            'radiate',
            'demand',
        ],
        'buy': ['purchase', 'acquire', 'order'],
        'sell': ['market', 'retail'],
        'earn': ['gain', 'receive', 'collect', 'accrue', 'credit'],
        'spend': ['pay', 'expense', 'debit'],
        'emit': ['emission', 'expel', 'pollute', 'pollution'],
        'abate': ['diminish', 'lessen', 'curtail', 'sequester'],
        'benefit': ['advantage', 'profit', 'reward', 'value'],
        'detriment': ['damage', 'harm', 'disadvantage'],
        'dispose': ['discard', 'eliminate', 'get_rid_of', 'dump'],
        'use': ['employ', 'apply', 'exploit'],
    }

    # makes aliases for key_aspect by transferring
    # right_aspect aliases onto left_aspect aliases
    # right_aspect_left_aspect is also an aliases
    transfer = {
        'invsetup': ('inventory', 'setup'),
        'invcapacity': ('inventory', 'capacity'),
        'invdismantle': ('inventory', 'dismantle'),
    }

    for key_aspect, (right_aspect, left_aspect) in transfer.items():

        # add key aspect
        _default_aliases[key_aspect] = [f'{right_aspect}_{left_aspect}']

        for right_alias in _default_aliases[right_aspect]:
            for left_alias in _default_aliases[left_aspect]:
                _default_aliases[key_aspect].append(f"{right_alias}_{left_alias}")

    return _default_aliases


default_aliases = _create_default_aliases()


def aspect_aliases(m: Model):
    """Define aspect aliases for the model.


    :param m:  The model instance to define aliases for.
    :type m: Model

    .. note::

        I have added what I could think off. Update this for reconfiguration.
        For just adding aliases use:
        model.aliases('alias1', ..., to='aspect') or model.aspect.aliases('alias', ...)

    """

    for aspect, aliases in default_aliases.items():
        m.alias(*aliases, of=aspect)


# FOR TESTING PURPOSES ONLY
# synonym_to_keys = {}
# for key, synonyms in default_aliases.items():
#     for syn in synonyms:
#         synonym_to_keys.setdefault(syn, []).append(key)

# # Print duplicates
# for syn, keys in synonym_to_keys.items():
#     if len(keys) > 1:
#         print(f"'{syn}' appears in multiple keys: {keys}")



"""Inherited _Name (reprs) class"""

from ._hash import _Hash


# Donot use this for any component that has a built-in name generating scheme
# Use Hash instead
class _Name(_Hash):
    """
    Inherited by components that only have a name.

    These do not feature as indices in the model
    nor do they have modeling aspects.

    For components with a:
        - mathematical program, use `X`
        - mathematical program and
          input parameters (modeling aspects), use `Component`

    :param label: Label used for plotting. Defaults to None.
    :type label: str, optional

    :ivar name: Name of the object. Defaults to ''.
    :vartype name: str

    .. note:
        - `name` is set when the component is made a Model attribute.
    """

    def __init__(self, label: str = ""):
        self.label = label
        self.name = ""
